// module hu_trace.cpp
// Dr.E.Huckert 02-2005

// Versions:
//  07-2006: two alternative implementations: file based with immediate
//           write-thru to file or
//           list based (with serialisation in the destructor)
//  02-2008: include Dateien ueberarbeitet
//  03-2008  logging via UDP added, list based logging dropped
//  10-2011 new method trace() with 2 int args
//  08-2014  two static functions added: get_time() and close_socket()

#include "hu_trace.h"
//#include "util_hu.h"

// ------------------------------------------------------  
// convert a binary IP address (IPV4) to a string
// Warning: this is not reentrant!
char * HU_TRACE::convert_binary_ip(char *binstr)
{
int n, ic;
char xbuf[8];
static char retbuf[32];
  retbuf[0] = 0;
  for (n=0; n < 4; n++)
  {
    ic = binstr[n] & 0xff;
    ::SPRINTF(xbuf, "%0d", ic);
    ::strcat(retbuf, xbuf);
    if (n < 3) ::strcat(retbuf, ".");
  }
  return retbuf;
}   // end convert_binary_ip()

// -------------------------------------------------------
int HU_TRACE::UDP_send(sock_t sock, 
                       const char *data,   // data to be sent out
                       int size,  // no. of bytes to send
                       const char *serverName,
                       int  port)
{
  struct sockaddr_in addr_sendto;
  struct hostent *h;
  int    rc;
  char   inetaddr[32];
  char   **pptr;
  char   *caddr;
  //
  // resolve the servername into an IP address
  h = ::gethostbyname(serverName);
  if (h == NULL)
  {
     return -1;
  }
  for (pptr = h->h_addr_list; *pptr != NULL; pptr++)
  {
    caddr = this->convert_binary_ip(*pptr); 
    ::strcpy(inetaddr, caddr);
  }
  ::memset((char *)&addr_sendto, 0, sizeof(addr_sendto));
  addr_sendto.sin_family      = AF_INET;
  addr_sendto.sin_port        = htons(port);
  addr_sendto.sin_addr.s_addr = ::inet_addr(inetaddr);
  //
  // send now
  rc = ::sendto(sock, data, (size_t)size, 0,
                (struct sockaddr *)&addr_sendto,
                sizeof(addr_sendto));
  return rc;
}   // end HU_TRACE::UDP_send()

// -------------------------------------------------------------------------
// Returns < 0 upon error
int HU_TRACE::UDP_send(const char *msg, int noBytes)
{
  int ret;
  ret = this->UDP_send(this->sock, msg, noBytes, 
                       this->udpServer, this->udpPort);
  return ret;  
}

// -------------------------------------------------------
sock_t HU_TRACE::create_socket( int af, int type, int protocol )
{
    sock_t sock;
    const int y = 1;
    sock = ::socket(af, type, protocol);
    if (sock < 0) return sock;
    ::setsockopt( sock, SOL_SOCKET,
                SO_REUSEADDR, (char *)(&y), sizeof(int));
    return sock;
}

// -----------------------------------------------------------------------
// liefert Zeit im Format SS:MM:SS:mmm 
static void get_time(char *dest)
{
  time_t   ltime;
  struct tm *newtime;
  struct timeb tb;
  //
  time(&ltime);  // 13.08.2009 gefaehrlich: 2 systems calls (real time!!!!)
  newtime = localtime(&ltime);
  ftime(&tb);    // 13.08.2009 gefaehrlich: 2 systems calls (real time!!!!)
  sprintf(dest,"%02d:%02d:%02d.%03d",
          (int)newtime->tm_hour,
          (int)newtime->tm_min,
          (int)newtime->tm_sec,
          (int)tb.millitm);
}   // end get_time()

// -------------------------------------------------------
static void close_socket(sock_t *sock)
{
  if (*sock != -1)
#ifdef WIN32
    ::closesocket(*sock);
#else
  ::close(*sock);
#endif
}

// -------------------------------------------------------
void HU_TRACE::cleanup(void)
{
#ifdef WIN32
   /* Cleanup Winsock */
   WSACleanup();
#endif
}   // end HU_TRACE::cleanup()

// ------------------------------------------------------------------------
void HU_TRACE::initialize(const char *fn,
                          const char *udpserver,
                          const int udpport)
{
  this->error = 0;
  #ifdef WIN32
  if (::crit_trace_initialized == 0)
  {
    ::crit_trace_initialized = 1;
    InitializeCriticalSection(&::crit_trace);
  }
  #endif
  this->setLock();
  if (fn != NULL)
  {
    // file based
    this->ofd = creat(fn, 0777);
    this->setFileName(fn);
    this->count = 0L;
    this->udpServer = NULL;
    this->udpPort   = -1;
    this->sock = -1;
  }
  else
  {
    // UDP based
    #ifdef WIN32
    if (WSAStartup(0x0101, &(WSAdata)) != 0)
    {
      this->error++;
      this->unLock();
      printf("ERROR 1 UDP logging\n"); 
      return;
    }
    #endif
    udpServer = new char[128];
    ::strcpy(this->udpServer, udpserver);
    this->udpPort = udpport;
    this->filename = NULL;
    this->ofd = -1;
    this->sock = this->create_socket(AF_INET, SOCK_DGRAM, 0);
    if (sock < 0)
    {
      this->error++;
      printf("ERROR 2 UDP logging\n"); 
    }
    //else
    //  printf("UDP logging ok\n"); 
  }
  this->unLock();	
} // end HU_TRACE::initialize()

// ------------------------------------------------------------------------
HU_TRACE::HU_TRACE()
{
  this->initialize("hu_trace.log", NULL, -1);
}

// ------------------------------------------------------------------------
// Simple constructor for file base logging
HU_TRACE::HU_TRACE(const char *fn)
{
   this->initialize(fn, NULL, -1);
}

// ------------------------------------------------------------------------
// simple constructor for UDP based logging
HU_TRACE::HU_TRACE(const char * udpserver, int udpport)
{
   this->initialize(NULL, udpserver, udpport);
}


// ------------------------------------------------------------------------
// simple constructor for UDP based logging
HU_TRACE::HU_TRACE(const char *idString,      // ignored
                   const char * udpserver, 
                   int udpport)
{
   this->initialize(NULL, udpserver, udpport);
}

// ------------------------------------------------------------------------
void HU_TRACE::setLock()
{
#ifdef WIN32
  EnterCriticalSection(&::crit_trace);
#else
#endif
}

// ------------------------------------------------------------------------
void HU_TRACE::unLock()
{
#ifdef WIN32
  LeaveCriticalSection(&::crit_trace);
#else
#endif
}

// ------------------------------------------------------------------------
// Destructor
HU_TRACE::~HU_TRACE()
{
  if (this->filename != NULL)
  {
    close(ofd);
    delete this->filename;
  }
  else
  {
    if (udpServer != NULL)
    {
      this->cleanup();
      delete udpServer;
    }
    udpServer = NULL;
  }
  this->count = 0L;
}

// ------------------------------------------------------------------------
void HU_TRACE::setFileName(const char *fn)
{
  this->filename = new char[strlen(fn) + 1];
  strcpy(this->filename, fn);
}

// ------------------------------------------------------------------------
// The is the central loggin routine - all other output routines end here!
// Assumes that the messages have normally no LF
void HU_TRACE::trace(const char *msg)
{
  this->setLock();
  int ret, ll;
  //
  if (msg == NULL)
    return;
  ll = ::strlen(msg) + 1;
  char *buf = new char[ll + 32];
  ::get_time(buf);
  ::strcat(buf, " ");
  ::strcat(buf, msg);
  ::strcat(buf, "\n");
  if (this->ofd >= 0)
  {
    // logging is file based
    if (msg != NULL)
    {
      this->count++;
      ret = ::write(this->ofd, buf, ::strlen(buf));
    }
  }
  else
  {
    // logging is UDP based
    ret = this->UDP_send(buf, ::strlen(buf));
    //printf("sent per UDP: ret=%d msg=%s\n", ret, msg);
  }
  if (buf != NULL)
    delete buf;
  this->unLock();
}   // end HU_TRACE::trace()

// ------------------------------------------------------------------------
// appends a LF at the end of the string to be written
void HU_TRACE::traceLF(const char *msg)
{
  if (msg != NULL)
  {
    char *aux;
    aux = new char[strlen(msg) + 4];
    strcpy(aux, msg);
    strcat(aux, "\n");
    this->trace(aux);
    delete aux;
  }
}

// ------------------------------------------------------------------------
void HU_TRACE::trace(const char *msg, const char *arg1)
{
  if ((msg != NULL) && (arg1 != NULL))
  {
    //wxString s1(msg);
    //s1.append(arg1);
    //this->trace(s1.c_str());
    int l1 = strlen(msg);
    int l2 = strlen(arg1);
    char *aux = new char[l1 + l2 + 10];
    strcpy(aux, msg);
    strcat(aux, arg1);
    this->trace(aux);
    delete aux;
  }
}   // end HU_TRACE::trace()

// ------------------------------------------------------------------------
void HU_TRACE::trace(const char *msg, int arg1)
{
  if (msg != NULL)
  {
    char *caddr;
    char aux[32];
    SPRINTF(aux, "%d", arg1);
    caddr = new char[strlen(msg) + strlen(aux) + 20];
    strcpy(caddr, msg);
    strcat(caddr, aux);
    this->trace(caddr);
    delete caddr;
  }
}   // end HU_TRACE::trace()

// ------------------------------------------------------------------------
void HU_TRACE::trace(const char *msg, int arg1, int arg2)
{
  if (msg != NULL)
  {
    char *caddr;
    char aux[48];
    SPRINTF(aux, "%d %d", arg1, arg2);
    caddr = new char[strlen(msg) + strlen(aux) + 20];
    strcpy(caddr, msg);
    strcat(caddr, aux);
    this->trace(caddr);
    delete caddr;
  }
}   // end HU_TRACE::trace()

// ------------------------------------------------------------------------
void HU_TRACE::trace(const char *msg, int arg1, int arg2, int arg3)
{
  if (msg != NULL)
  {
    char *caddr;
    char aux[64];
    SPRINTF(aux, "%d %d %d", arg1, arg2, arg3);
    caddr = new char[strlen(msg) + strlen(aux) + 20];
    strcpy(caddr, msg);
    strcat(caddr, aux);
    this->trace(caddr);
    delete caddr;
  }
}   // end HU_TRACE::trace()

// ------------------------------------------------------------------------
void HU_TRACE::trace(const char *msg, int arg1, int arg2, int arg3, int arg4)
{
  if (msg != NULL)
  {
    char *caddr;
    char aux[96];
    SPRINTF(aux, "%d %d %d %d", arg1, arg2, arg3, arg4);
    caddr = new char[strlen(msg) + strlen(aux) + 20];
    strcpy(caddr, msg);
    strcat(caddr, aux);
    this->trace(caddr);
    delete caddr;
  }
}   // end HU_TRACE::trace()

// ------------------------------------------------------------------------
void HU_TRACE::trace(const char *msg, long arg1)
{
  if (msg != NULL)
  {
    char *caddr;
    char aux[32];
    SPRINTF(aux, "%ld", arg1);
    caddr = new char[strlen(msg) + strlen(aux) + 20];
    strcpy(caddr, msg);
    strcat(caddr, aux);
    this->trace(caddr);
    delete caddr;
  }
}   // end HU_TRACE::trace()

// ------------------------------------------------------------------------
void HU_TRACE::trace(const char *msg, float arg1)
{
  if (msg != NULL)
  {
    char *caddr;
    char aux[32];
    ::SPRINTF(aux, "%f", arg1);
    caddr = new char[strlen(msg) + strlen(aux) + 20];
    strcpy(caddr, msg);
    strcat(caddr, aux);
    this->trace(caddr);
    delete caddr;
  }
}   // end HU_TRACE::trace()

// ------------------------------------------------------------------------
void HU_TRACE::trace(const char *msg, unsigned int arg1)
{
  if (msg != NULL)
  {
    char *caddr;
    char aux[32];
    ::SPRINTF(aux, "%u", arg1);
    caddr = new char[strlen(msg) + strlen(aux) + 20];
    strcpy(caddr, msg);
    strcat(caddr, aux);
    this->trace(caddr);
    delete caddr;
  }
}   // end HU_TRACE::trace()

// ------------------------------------------------------------------------
void HU_TRACE::trace(const char *msg, unsigned long arg1)
{
  if (msg != NULL)
  {
    char *caddr;
    char aux[32];
    ::SPRINTF(aux, "%lu", arg1);
    caddr = new char[strlen(msg) + strlen(aux) + 20];
    strcpy(caddr, msg);
    strcat(caddr, aux);
    this->trace(caddr);
    delete caddr;
  }
}   // end HU_TRACE::trace()

// ------------------------------------------------------------------------
void HU_TRACE::tracex(const char *msg, long arg1)
{
  if (msg != NULL)
  {
    char *caddr;
    char aux[32];
    ::SPRINTF(aux, "%08lx", arg1);
    caddr = new char[strlen(msg) + strlen(aux) + 20];
    strcpy(caddr, msg);
    strcat(caddr, aux);
    this->trace(caddr);
    delete caddr;
  }
}   // end HU_TRACE::trace()

// ------------------------------------------------------------------------
void HU_TRACE::trace(const char *msg, const char arg1)
{
  if (msg != NULL)
  {
    char *caddr;
    char aux[32];
    ::SPRINTF(aux, "%c", arg1);
    caddr = new char[strlen(msg) + strlen(aux) + 20];
    strcpy(caddr, msg);
    strcat(caddr, aux);
    this->trace(caddr);
    delete caddr;
  }
}   // end HU_TRACE::trace()

// ------------------------------------------------------------------------
void HU_TRACE::trace(const STRING msg)
{
  this->trace((const char *)(msg.c_str()));
}

// ------------------------------------------------------------------------
void HU_TRACE::trace(const char *p1, const STRING p2, int p3)
{
  STRING s;
  s = p1;
  s = s + p2 + " ";
  this->trace(s.c_str(), p3);
}

// ------------------------------------------------------------------------
void HU_TRACE::trace(const STRING *pmsg)
{
  if (pmsg != NULL)
  {
    this->trace((const char *)(pmsg->c_str()));
  }
}

// ------------------------------------------------------------------------
void HU_TRACE::trace(const char *pstr, const STRING msg)
{
  if (pstr != NULL)
  {
    this->trace(pstr, (const char *)(msg.c_str()));
  }
}

// ------------------------------------------------------------------------
void HU_TRACE::trace(const char *pstr, const STRING *pmsg)
{
  if (pstr != NULL)
  {
    this->trace(pstr, (const char *)(pmsg->c_str()));
  }
}

